import asyncio

from . import _task_pool

__all__ = ["observable"]


class observable:
    def __init__(
        self,
        task_pool: _task_pool.task_pool,
    ):
        self.observers = []
        self._task_pool = task_pool

    def add_observer(self, function):
        self.observers.append(function)

    def observer(self, function):
        self.add_observer(function)
        return function

    def notify(self, *args, **kwargs):
        for observer in self.observers:
            # FIXME
            maybe_coroutine = observer(*args, **kwargs)
            if asyncio.iscoroutine(maybe_coroutine):
                self._task_pool.schedule(maybe_coroutine)



from dataclasses import dataclass, field

from ._decoding import *
from ._encoding import *
from ._new_id import *
from ._observable import *
from ._schema import *
from ._streaming import *
from ._task_pool import *
from ._validate_execution import *
from ._is_valid_uri import *

__all__ = [
    "dataclass",
    "field",
    *_decoding.__all__,
    *_encoding.__all__,
    *_new_id.__all__,
    *_observable.__all__,
    *_schema.__all__,
    *_streaming.__all__,
    *_task_pool.__all__,
    *_validate_execution.__all__,
    *_is_valid_uri.__all__,
]



import asyncio
import typing

from . import _almanet
from . import _shared

__all__ = [
    "microservice",
]


@_shared.dataclass
class abstract_procedure_model:
    microservice: "microservice"
    procedure: typing.Callable
    label: str = ...
    channel: str | None = None
    include_to_api: bool = True
    description: str | None = None
    tags: set[str] = _shared.field(default_factory=set)
    validate: bool = True
    payload_model: typing.Any = ...
    return_model: typing.Any = ...

    @property
    def uri(self):
        return self.microservice._make_uri(self.label)

    def __post_init__(self):
        if not callable(self.procedure):
            raise ValueError("decorated function must be callable")
        if self.label is ...:
            self.label = self.procedure.__name__
        if self.description is None:
            self.description = self.procedure.__doc__
        self.payload_model, self.return_model = _shared.extract_annotations(
            self.procedure,
            self.payload_model,
            self.return_model
        )

    def implementation(
        self,
        real_procedure: typing.Callable,
    ) -> _almanet.registration_model:
        return self.microservice.register_procedure(
            real_procedure,
            label=self.label,
            channel=self.channel,
            include_to_api=self.include_to_api,
            description=self.description,
            tags=self.tags,
            validate=self.validate,
            payload_model=self.payload_model,
            return_model=self.return_model,
        )


class microservice:
    """
    Represents a microservice that can be used to register procedures (functions) with a session.
    """

    class _kwargs(typing.TypedDict):
        """
        - prepath: is used to prepend a path to the procedure's topic.
        - tags: are used to categorize the procedures.
        """

        prepath: typing.NotRequired[str]
        tags: typing.NotRequired[typing.Set[str]]

    def __init__(
        self,
        session: _almanet.Almanet,
        **kwargs: typing.Unpack[_kwargs],
    ):
        self._routes = set()
        self.pre = kwargs.get("prepath")
        self.tags = set(kwargs.get("tags") or [])
        self.session = session

    def _share_self_schema(
        self,
        **extra,
    ) -> None:
        async def procedure(*args, **kwargs):
            return {
                "client": self.session.id,
                "version": self.session.version,
                "routes": list(self._routes),
                **extra,
            }

        self.session.register(
            "_api_schema_.client",
            procedure,
            channel=self.session.id,
        )

    def _share_procedure_schema(
        self,
        uri: str,
        channel: str,
        tags: set[str] | None = None,
        **extra,
    ) -> None:
        if tags is None:
            tags = set()
        tags |= self.tags
        if len(tags) == 0:
            tags = {"Default"}

        async def procedure(*args, **kwargs):
            return {
                "client": self.session.id,
                "version": self.session.version,
                "uri": uri,
                "channel": channel,
                "tags": tags,
                **extra,
            }

        self.session.register(
            f"_api_schema_.{uri}.{channel}",
            procedure,
            channel=channel,
        )

        self._routes.add(f"{uri}/{channel}")

    def _make_uri(
        self,
        subtopic: str,
    ) -> str:
        return f"{self.pre}.{subtopic}" if isinstance(self.pre, str) else subtopic

    class _register_procedure_kwargs(typing.TypedDict):
        label: typing.NotRequired[str]
        channel: typing.NotRequired[str | None]
        include_to_api: typing.NotRequired[bool]
        description: typing.NotRequired[str | None]
        tags: typing.NotRequired[set[str]]
        validate: typing.NotRequired[bool]
        payload_model: typing.NotRequired[typing.Any]
        return_model: typing.NotRequired[typing.Any]

    def register_procedure(
        self,
        procedure: typing.Callable,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> "_almanet.registration_model":
        if not callable(procedure):
            raise ValueError("decorated function must be callable")

        label = kwargs.pop("label", procedure.__name__)
        uri = self._make_uri(label)

        payload_model = kwargs.pop("payload_model", ...)
        return_model = kwargs.pop("return_model", ...)
        if kwargs.get("validate", True):
            procedure = _shared.validate_execution(procedure, payload_model, return_model)

        registration = self.session.register(uri, procedure, channel=kwargs.pop("channel", None))

        if kwargs.get("include_to_api", True):
            procedure_schema = _shared.describe_function(
                procedure,
                kwargs.pop("description"),
                payload_model,
                return_model,
            )
            self._share_procedure_schema(
                uri,
                registration.channel,
                **kwargs,  # type: ignore
                **procedure_schema,  # type: ignore
            )

        return registration

    def procedure(
        self,
        function: typing.Callable | None = None,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ):
        """
        Allows you to easily add procedures (functions) to a microservice by using a decorator.
        Returns a decorated function.
        """
        if function is None:
            return lambda function: self.register_procedure(function, **kwargs)
        return self.register_procedure(function, **kwargs)

    def abstract_procedure(
        self,
        function: typing.Callable | None = None,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ):
        if function is None:
            return lambda function: abstract_procedure_model(self, function, **kwargs)
        return abstract_procedure_model(self, function, **kwargs)

    def serve(self) -> None:
        """
        Runs an event loop to serve the microservice.
        """
        self.session._post_join_event.add_observer(self._share_self_schema)

        loop = asyncio.new_event_loop()
        loop.create_task(self.session.join())
        loop.run_forever()



import asyncio
import logging
import typing

import pydantic_core

from . import _shared

if typing.TYPE_CHECKING:
    from . import _microservice

__all__ = [
    "Almanet",
    "client_iface",
    "invoke_event_model",
    "qmessage_model",
    "reply_event_model",
    "rpc_error",
]


logger = logging.getLogger("almanet")
logger.setLevel(logging.INFO)


@_shared.dataclass(slots=True)
class qmessage_model[T: typing.Any]:
    """
    Represents a message in the queue.
    """

    id: str
    timestamp: int
    body: T
    attempts: int
    commit: typing.Callable[[], typing.Awaitable[None]]
    rollback: typing.Callable[[], typing.Awaitable[None]]


type returns_consumer[T: typing.Any] = tuple[typing.AsyncIterable[qmessage_model[T]], typing.Callable[[], None]]


class client_iface(typing.Protocol):
    """
    Interface for a client library.
    """

    async def connect(
        self,
        addresses: typing.Sequence[str],
    ) -> None:
        raise NotImplementedError()

    async def produce(
        self,
        topic: str,
        message: str | bytes,
    ) -> None:
        raise NotImplementedError()

    async def consume(
        self,
        topic: str,
        channel: str,
    ) -> returns_consumer[bytes]:
        raise NotImplementedError()

    async def close(self) -> None:
        raise NotImplementedError()


@_shared.dataclass(slots=True)
class invoke_event_model[T: typing.Any]:
    """
    Represents an invocation event.
    """

    id: str
    caller_id: str
    payload: T
    reply_topic: str

    @property
    def expired(self) -> bool:
        # TODO
        return False


@_shared.dataclass(slots=True)
class reply_event_model[T: typing.Any]:
    """
    Represents a reply event.
    """

    call_id: str
    is_error: bool
    payload: T


class rpc_error(Exception):
    """
    Represents an RPC error.
    You can inherit from this class to create your own error.
    """

    __slots__ = ("name", "args")

    def __init__(
        self,
        *args,
        name: str | None = None,
    ) -> None:
        self.name = name or self.__class__.__name__
        self.args = args

    def __str__(self) -> str:
        return f"{self.name}{self.args}"


@_shared.dataclass(slots=True)
class registration_model:
    """
    Represents a registered procedure to call.
    """

    uri: str
    channel: str
    procedure: typing.Callable
    session: "Almanet"

    @property
    def __name__(self):
        return self.uri

    @property
    def __doc__(self):
        return self.procedure.__doc__

    @property
    def __call__(self):
        return self.procedure

    async def execute(
        self,
        invocation: invoke_event_model,
    ) -> reply_event_model:
        __log_extra = {"registration": str(self), "invocation": str(invocation)}
        try:
            logger.debug("trying to execute procedure", extra=__log_extra)
            if asyncio.iscoroutinefunction(self.procedure):
                reply_payload = await self.procedure(invocation.payload)
            else:
                reply_payload = await asyncio.to_thread(self.procedure, invocation.payload)
            return reply_event_model(call_id=invocation.id, is_error=False, payload=reply_payload)
        except Exception as e:
            if isinstance(e, rpc_error):
                error_name = e.name
                error_message = e.args
            elif isinstance(e, pydantic_core.ValidationError):
                error_name = "ValidationError"
                error_message = repr(e)
            else:
                error_name = "InternalError"
                error_message = "oops"
                logger.exception("during execute procedure", extra=__log_extra)
            return reply_event_model(
                call_id=invocation.id,
                is_error=True,
                payload={"name": error_name, "message": error_message},
            )


class Almanet:
    """
    Represents a session, connected to message broker.
    """

    @property
    def version(self) -> float:
        return 0

    class _kwargs(typing.TypedDict):
        id: typing.NotRequired[str]

    def __init__(
        self,
        *addresses: str,
        client: client_iface,
        **kwargs: typing.Unpack[_kwargs],
    ) -> None:
        if not all(isinstance(i, str) for i in addresses):
            raise ValueError("addresses must be a iterable of strings")
        self.id = kwargs.get("id") or _shared.new_id()
        self.joined = False
        self.addresses = addresses
        self._client = client
        self.task_pool = _shared.task_pool()
        self._post_join_event = _shared.observable(self.task_pool)
        self._leave_event = _shared.observable(self.task_pool)
        self.__pending_replies: typing.MutableMapping[str, asyncio.Future[reply_event_model]] = {}

    type __produce_args = tuple[str, typing.Any]

    async def __produce(
        self,
        topic: str,
        payload: typing.Any,
    ) -> None:
        try:
            message_body = _shared.dump(payload)
        except Exception as e:
            logger.error(f"during encode payload: {repr(e)}")
            raise e

        try:
            logger.debug(f"trying to produce {topic} topic")
            await self._client.produce(topic, message_body)
        except Exception as e:
            logger.exception(f"during produce {topic} topic")
            raise e

    def produce(
        self,
        *args: typing.Unpack[__produce_args],
    ) -> asyncio.Task[None]:
        """
        Produce a message with a specified topic and payload.
        """
        return self.task_pool.schedule(self.__produce(*args))

    async def _serialize[T: typing.Any](
        self,
        messages_stream: typing.AsyncIterable[qmessage_model[bytes]],
        payload_model: type[T] | typing.Any = ...,
    ) -> typing.AsyncIterable[qmessage_model[T]]:
        serializer = _shared.serialize_json(payload_model)

        async for message in messages_stream:
            try:
                message.body = serializer(message.body)
            except:
                logger.exception("during decode payload")
                continue

            yield message  # type: ignore

    async def consume[T: typing.Any](
        self,
        topic: str,
        channel: str,
        *,
        payload_model: type[T] | typing.Any = ...,
    ) -> returns_consumer[T]:
        """
        Consume messages from a message broker with the specified topic and channel.
        It returns a tuple of a stream of messages and a function that can stop consumer.
        """
        logger.debug(f"trying to consume {topic}/{channel}")

        messages_stream, stop_consumer = await self._client.consume(topic, channel)
        self._leave_event.add_observer(stop_consumer)

        messages_stream = self._serialize(messages_stream, payload_model)

        return messages_stream, stop_consumer

    async def _consume_replies(
        self,
        ready_event: asyncio.Event,
    ) -> None:
        messages_stream, _ = await self.consume(
            f"_rpc_._reply_.{self.id}",
            channel="rpc-recipient",
        )
        logger.debug("reply event consumer begin")
        ready_event.set()
        async for message in messages_stream:
            __log_extra = {"incoming_message": str(message)}
            try:
                reply = reply_event_model(**message.body)
                __log_extra["reply"] = str(reply)
                logger.debug("new reply", extra=__log_extra)

                pending = self.__pending_replies.get(reply.call_id)
                if pending is None:
                    logger.warning("pending event not found", extra=__log_extra)
                else:
                    pending.set_result(reply)
            except:
                logger.exception("during parse reply", extra=__log_extra)

            await message.commit()
            logger.debug("successful commit", extra=__log_extra)
        logger.debug("reply event consumer end")

    type __call_args = tuple[typing.Any, typing.Any]

    class __call_kwargs(typing.TypedDict):
        timeout: typing.NotRequired[int]
        return_model: typing.NotRequired[typing.Any]

    async def __call(
        self,
        topic: typing.Union[str, "_microservice.abstract_procedure_model"],
        payload: typing.Any,
        *,
        timeout: int = 60,
        return_model: typing.Any = ...,
    ) -> reply_event_model:
        if not isinstance(topic, str):
            if return_model is ...:
                return_model = topic.return_model
            topic = topic.uri

        if return_model is ...:
            serialize_return = lambda x: x
        else:
            serialize_return = _shared.serialize(return_model)

        invocation = invoke_event_model(
            id=_shared.new_id(),
            caller_id=self.id,
            payload=payload,
            reply_topic=f"_rpc_._reply_.{self.id}",
        )

        __log_extra = {"topic": topic, "timeout": timeout, "invoke_event": str(invocation)}
        logger.debug("trying to call", extra=__log_extra)

        pending_reply_event = asyncio.Future[reply_event_model]()
        self.__pending_replies[invocation.id] = pending_reply_event

        try:
            async with asyncio.timeout(timeout):
                await self.produce(f"_rpc_.{topic}", invocation)

                response = await pending_reply_event
                __log_extra["reply_event"] = str(response)
                logger.debug("new reply event", extra=__log_extra)

                if response.is_error:
                    raise rpc_error(
                        response.payload["message"],
                        name=response.payload["name"],
                    )

                response.payload = serialize_return(response.payload)

                return response
        except Exception as e:
            logger.error("during call", extra={**__log_extra, "error": repr(e)})
            raise e
        finally:
            self.__pending_replies.pop(invocation.id)

    def call(
        self,
        *args: typing.Unpack[__call_args],
        **kwargs: typing.Unpack[__call_kwargs],
    ) -> asyncio.Task[reply_event_model]:
        """
        Call a procedure with a specified topic and payload.
        Returns a reply event.
        """
        return self.task_pool.schedule(self.__call(*args, **kwargs))

    type __multicall_args = tuple[str, typing.Any]

    class __multicall_kwargs(typing.TypedDict):
        timeout: typing.NotRequired[int]

    async def __multicall(
        self,
        topic: typing.Union[str, "_microservice.abstract_procedure_model"],
        payload: typing.Any,
        *,
        timeout: int = 60,
    ) -> list[reply_event_model]:
        if not isinstance(topic, str):
            topic = topic.uri

        invocation = invoke_event_model(
            id=_shared.new_id(),
            caller_id=self.id,
            payload=payload,
            reply_topic=f"_rpc_._replies_.{self.id}",
        )

        __log_extra = {"topic": topic, "timeout": timeout, "invoke_event": str(invocation)}

        messages_stream, stop_consumer = await self.consume(invocation.reply_topic, "rpc-recipient")

        result = []
        try:
            async with asyncio.timeout(timeout):
                await self.produce(f"_rpc_.{topic}", invocation)

                async for message in messages_stream:
                    try:
                        logger.debug("new reply event", extra=__log_extra)
                        reply = reply_event_model(**message.body)
                        result.append(reply)
                    except:
                        logger.exception("during parse reply event", extra=__log_extra)

                    await message.commit()
        except TimeoutError:
            stop_consumer()

        logger.debug(f"multicall {topic} done")

        return result

    async def multicall(
        self,
        *args: typing.Unpack[__multicall_args],
        **kwargs: typing.Unpack[__multicall_kwargs],
    ) -> asyncio.Task[list[reply_event_model]]:
        """
        Call simultaneously multiple procedures with a specified topic and payload.
        Returns a list of reply events.
        """
        return self.task_pool.schedule(self.__multicall(*args, **kwargs))

    async def _consume_invocations(
        self,
        registration: registration_model,
    ) -> None:
        logger.debug(f"trying to register {registration.uri}/{registration.channel}")
        messages_stream, _ = await self.consume(f"_rpc_.{registration.uri}", registration.channel)
        async for message in messages_stream:
            __log_extra = {"registration": str(registration), "incoming_message": str(message)}
            try:
                invocation = invoke_event_model(**message.body)
                __log_extra["invocation"] = str(invocation)
                logger.debug("new invocation", extra=__log_extra)

                if invocation.expired:
                    logger.warning("invocation expired", extra=__log_extra)
                else:
                    reply = await registration.execute(invocation)
                    logger.debug("trying to reply", extra=__log_extra)
                    await self.produce(invocation.reply_topic, reply)
            except:
                logger.exception("during execute invocation", extra=__log_extra)

            await message.commit()
            logger.debug("successful commit", extra=__log_extra)
        logger.debug(f"consumer {registration.uri} down")

    def register(
        self,
        topic: str,
        procedure: typing.Callable,
        *,
        channel: str | None = None,
    ) -> registration_model:
        """
        Register a procedure with a specified topic and payload.
        Returns the created registration.
        """
        r = registration_model(
            uri=topic,
            channel=channel or "RPC",
            procedure=procedure,
            session=self,
        )

        self._post_join_event.add_observer(lambda: self._consume_invocations(r))

        return r

    async def join(self) -> None:
        """
        Join the session to message broker.
        """
        if self.joined:
            raise RuntimeError(f"session {self.id} already joined")

        logger.debug(f"trying to connect addresses={self.addresses}")

        await self._client.connect(self.addresses)

        consume_replies_ready = asyncio.Event()
        self.task_pool.schedule(
            self._consume_replies(consume_replies_ready),
            daemon=True,
        )
        await consume_replies_ready.wait()

        self.joined = True
        self._post_join_event.notify()
        logger.info(f"session {self.id} joined")

    async def __aenter__(self) -> "Almanet":
        if not self.joined:
            await self.join()
        return self

    async def leave(
        self,
        reason: str | None = None,
    ) -> None:
        """
        Leave the session from message broker.
        """
        if not self.joined:
            raise RuntimeError(f"session {self.id} not joined")

        self.joined = False

        logger.debug(f"trying to leave {self.id} session, reason: {reason}")

        stop_consume_replies = self._leave_event.observers.pop(0)

        self._leave_event.notify()

        logger.debug(f"session {self.id} await task pool complete")
        await self.task_pool.complete()

        stop_consume_replies()

        logger.debug(f"session {self.id} trying to close connection")
        await self._client.close()

        logger.warning(f"session {self.id} left")

    async def __aexit__(
        self,
        exception_type,
        exception_value,
        exception_traceback,
    ) -> None:
        if self.joined:
            await self.leave()



import typing
from ._almanet import *
from . import _clients as clients
from ._flow import *
from ._microservice import *

__all__ = [
    *_almanet.__all__,
    "clients",
    "new_session",
    *_flow.__all__,
    *_microservice.__all__,
    "new_microservice",
]


def new_session(
    *addresses: str,
    client_klass: type[client_iface] = clients.DEFAULT_CLIENT,
    **kwargs: typing.Unpack[Almanet._kwargs],
) -> Almanet:
    """
    Returns a new instance of Almanet session.
    """
    return Almanet(*addresses, client=client_klass(), **kwargs)


def new_microservice(
    *addresses: str,
    session: Almanet | None = None,
    **kwargs: typing.Unpack[microservice._kwargs],
) -> microservice:
    """
    Returns a new instance of microservice.
    """
    if session is None:
        session = new_session(*addresses)
    return microservice(session, **kwargs)



import functools
import inspect

from . import _decoding
from . import _schema

__all__ = ["validate_execution"]


def validate_execution(
    function,
    payload_model=...,
    return_model=...,
):
    """
    Takes a function as input and returns a decorator.
    The decorator validates the input payload and output return of the function based on their annotations.

    Args:
    - function: the function to decorate with validator
    - payload_model: the model of the input
    - return_model: the model of the output
    """
    payload_model, return_model = _schema.extract_annotations(function, payload_model, return_model)

    def dont_validate(v):
        return v

    payload_validator = dont_validate if payload_model is ... else _decoding.serialize(payload_model)

    return_validator = dont_validate if return_model is ... else _decoding.serialize(return_model)

    @functools.wraps(function)
    async def async_decorator(payload, *args, **kwargs):
        payload = payload_validator(payload)
        result = await function(payload, *args, **kwargs)
        return return_validator(result)

    if inspect.iscoroutinefunction(function):
        return async_decorator

    @functools.wraps(function)
    def decorator(payload, *args, **kwargs):
        payload = payload_validator(payload)
        result = function(payload, *args, **kwargs)
        return return_validator(result)

    return decorator
